#! /usr/bin/env python
# -*- coding: utf-8 -*-

import math
from .. import imatrix
from numpy import *

#TODO:存了两份 考虑怎么合并
SubjectIndexs = {
	"count": 0,
	"play": 1,
	"replay": 2,
	"conv": 3
}

WorksIndexs = {
	"play": 0,
	"like": 1,
	"shareF": 2,
	"shareI": 3,
	"replay": 4,
	"longTime": 5,
	"subAdd": 6,
	"playTimeAvg": 7,
	"conv": 8,
	"duration": 9,
	"size": 10
}

SubjectTotalWeight = 0.3

SubjectWeight = {
	'count': 20,
	'play': 1,
	'replay': 5
}

WorksWeight = {
	'play': 1,
	'like': 10,
	'shareF': 15,
	'shareI': 20,
	'replay': 5,
	'longTime': 10,
	'subAdd': 20
}

TimeDecay = {
	'half': 168,
	'flutter': 48.0
}

SameAidDecay = 0.0

def calSubjectScore(subjectList):
	subjectArgs = []
	for subject in subjectList:
		subjectArgs.append([subject.count, subject.play, subject.replay])
	subjectMat = imatrix.inOne(mat(subjectArgs))

	for index, subject in enumerate(subjectList):
		subject.score = (subjectMat[index, 0] * SubjectWeight['count'] + subjectMat[index, 1] * SubjectWeight['play'] + subjectMat[index, 2] * SubjectWeight['replay']) * subject.conv

def calWorksScore(worksList, dauFactor=10):
	if len(worksList) == 0:
		return

	worksArgs = []
	worksDurationTimeAvg = {}
	worksDurationPlay = {}
	for works in worksList:
		worksArgs.append([works.play, works.like, works.shareF, works.shareI, works.replay, works.longPlay, works.sub])
		if worksDurationTimeAvg.has_key(works.durationIndex):
			worksDurationTimeAvg[works.durationIndex] += works.timeAvg
			worksDurationPlay[works.durationIndex] += works.play
		else:
			worksDurationTimeAvg[works.durationIndex] = works.timeAvg
			worksDurationPlay[works.durationIndex] = works.play
	worksMat = imatrix.inOne(mat(worksArgs))
	worksMatTwo = imatrix.inOneDiv(hstack((worksMat[:, 0] * WorksWeight['play'] + worksMat[:, 1] * WorksWeight['like'] + worksMat[:, 2] * WorksWeight['shareF'] + worksMat[:, 3] * WorksWeight['shareI'] + worksMat[:, 4] * WorksWeight['replay'], worksMat[:, 5] * WorksWeight['longTime'] + worksMat[:, 6] * WorksWeight['subAdd'])))
	worksMatFinal = imatrix.inOneDiv(worksMatTwo[:, 0] * (11 - dauFactor) + worksMatTwo[:, 1] * dauFactor)

	for index, works in enumerate(worksList):
		works.score = worksMatFinal[index, 0]
		works.ratio = worksDurationPlay[works.durationIndex]

def updateWeight(weight):
	global SubjectTotalWeight, SubjectWeight, WorksWeight, TimeDecay, SameAidDecay
	SubjectTotalWeight = weight['subjectTotal']
	SubjectWeight = weight['subject']
	WorksWeight = weight['works']
	TimeDecay = weight['timeDecay']
	SameAidDecay = weight['sameAidDecay'] if weight.has_key('sameAidDecay') else 0.0	

def calHotV2(subjectHistory, worksHistory, worksList, dauFactor, aidWeights, vidWeights):
	calWorksScore(worksList, 11 - dauFactor)

	baseHot = []
	for works in worksList:
		vid = works.vid
		aid = works.aid
		baseHot.append([subjectHistory[aid]['score'] * worksHistory[vid]['ratio'] if subjectHistory.has_key(aid) and worksHistory.has_key(vid) else 0])

	baseHotMat = imatrix.inOneDiv(mat(baseHot))

	hotList = {}
	for index, works in enumerate(worksList):
		vid = works.vid
		aid = works.aid
		actionHot = (works.score + (worksHistory[vid]['score'] if worksHistory.has_key(vid) else 0)) / 2
		if vidWeights.has_key(vid):
			weight = float(vidWeights[vid])
		elif aidWeights.has_key(int(aid)):
			weight = float(aidWeights[int(aid)])
		else:
			weight = 1
		hotList[vid] = {'vid': vid, 'aid': aid, 'title': works.title, 'hot': (baseHotMat[index, 0] * SubjectTotalWeight + actionHot) * getDecayRatio(works.publishTime) * weight * 100000, 'isNew': works.publishTime < 72}

	hotListFinal = {}
	for index, item in enumerate(sorted(hotList.items(), key=lambda x: x[1]['hot'], reverse=True)):
		item[1]['ranking'] = index + 1
		hotListFinal[item[0]] = item[1]

	return hotListFinal











def calHot(dauFactor, newShowMax, subjectDict, worksDict):
	subjectAids, subjectHotMat = calSubjectHot(subjectDict)
	worksVids, worksAids, worksHotMat, worksRate = calWorksHot(worksDict, dauFactor, newShowMax)

	shArray = []
	for index, aid in enumerate(worksAids):
		shArray.append([subjectHotMat[subjectAids.index(aid), 0] * worksDict[worksVids[index]]['durationAvgPlay']])

	shm = imatrix.inOneDiv(mat(shArray))
	whm = imatrix.inOneDiv(worksHotMat)

	#worksNew = {}
	for index, vid in enumerate(worksVids):
		baseHot = shm[index, 0] * SubjectTotalWeight #* len(subjectDict) / len(worksDict)
		actionHot = whm[index, 0]

		works = worksDict[vid]

		decayRatio = getDecayRatio(works['publishTime'])

		works['vid'] = vid
		works['totalHot'] = (baseHot + actionHot) * decayRatio
		works['baseHot'] = baseHot
		works['actionHot'] = actionHot
		works['decayRatio'] = decayRatio
		works['baseRate'] = worksRate[index, 0]

		works['isNew'] = works['publishTime'] < 72 and works['show'] <= newShowMax

	mHotList = {}
	aidDecays = {}
	for item in sorted(worksDict.items(), key=lambda x: x[1]['totalHot'], reverse=True):
		if index >= 50000:
			break
		works = item[1]
		aid = works['aid']
		if aidDecays.has_key(aid) and aidDecays[aid] < 0.25:
			aidDecays[aid] += SameAidDecay
		else:
			aidDecays[aid] = 0
		works['hot'] = int((1 - aidDecays[aid]) * works['totalHot'] * 100000) + calLowerBound(works['recShowPlay'], works['recShow'])
		mHotList[item[0]] = works

	hotList = {}
	for index, item in enumerate(sorted(mHotList.items(), key=lambda x: x[1]['hot'], reverse=True)):
		item[1]['ranking'] = index + 1
		hotList[item[0]] = item[1]
		'''
		for i in range(500):
			info = item[1].copy()
			info['vid'] = item[0] + str(i)
			hotList[item[0] + str(i)] = info
		'''

	#if hotList.has_key('5fe50cdd8028734584f0832d5338bfe5'):
	#	hotList['5fe50cdd8028734584f0832d5338bfe5']['hot'] = 200000

	return hotList

def calLowerBound(pos, n):
	if n == 0:
		return 0
	z = 1.96
	pos = min(pos, n)
	phat = 1.0 * pos / n
	return (phat + z * z/(2 * n)  -  z * math.sqrt((phat * (1 - phat) + z * z/(4 * n)) / n))/(1 + z * z / n)

def calSubjectHot(subjectDict):
	subjectAids = []
	subjectArgs = []
	for aid, subject in subjectDict.items():
		subjectAids.append(aid)
		subjectArgs.append(subject['statis'])
	subjectMat = imatrix.inOne(mat(subjectArgs))
	subjectHotMat = multiply((subjectMat[:, SubjectIndexs['count']] * SubjectWeight['count'] + subjectMat[:, SubjectIndexs['play']] * SubjectWeight['play'] + subjectMat[:, SubjectIndexs['replay']] * SubjectWeight['replay']), subjectMat[:, SubjectIndexs['conv']])
	return subjectAids, subjectHotMat

def calWorksHot(worksDict, dauFactor, newShowMax):
	worksVids = []
	worksAids = []
	worksArgs = []
	worksForceCredible = []
	for vid, works in worksDict.items():
		worksVids.append(vid)
		worksAids.append(works['aid'])
		worksArgs.append(works['statis'])
		worksForceCredible.append([1 if works['show'] > newShowMax else 0])
	worksMat = mat(worksArgs)
	worksBaseMat = imatrix.inOne(worksMat[:, [WorksIndexs['play'], WorksIndexs['like'], WorksIndexs['shareF'], WorksIndexs['shareI'], WorksIndexs['replay'], WorksIndexs['longTime'], WorksIndexs['subAdd']]])

	worksActionMat = worksBaseMat[:, WorksIndexs['play']] * WorksWeight['play'] + worksBaseMat[:, WorksIndexs['like']] * WorksWeight['like'] + worksBaseMat[:, WorksIndexs['shareF']] * WorksWeight['shareF'] + worksBaseMat[:, WorksIndexs['shareI']] * WorksWeight['shareI'] + worksBaseMat[:, WorksIndexs['replay']] * WorksWeight['replay']
	worksForceMat = multiply(multiply(multiply(worksBaseMat[:, WorksIndexs['longTime']] * WorksWeight['longTime'] + worksBaseMat[:, WorksIndexs['subAdd']] * WorksWeight['subAdd'], worksMat[:, WorksIndexs['playTimeAvg']]), worksMat[:, WorksIndexs['conv']]), mat(worksForceCredible))

	return worksVids, worksAids, worksActionMat * dauFactor + worksForceMat * (11 - dauFactor), getWorksRate(worksMat[:, [WorksIndexs['duration'], WorksIndexs['play']]])

def getDecayRatio(publishTime):
	return 1 - 1 / (1 + math.exp((TimeDecay['half'] - publishTime) / TimeDecay['flutter']))

def getWorksRate(mat):
	#TODO::目前还没有统计size
	total = mat[:,1].sum(axis=0)[0,0]
	avg = multiply(mat[:,0], mat[:,1]).sum(axis=0)[0,0] / total
	vari = sqrt(multiply(square(mat[:,0] - avg), mat[:,1]).sum(axis=0)[0,0] / total)
	return exp(-square(mat[:,0] - avg) / 2 / square(vari)) * (1 / math.sqrt(2 * math.pi) / vari)

'''

def calHot(subjectDict, worksDict, dau):
	worksVid = []
	worksAid = []
	matArray = []
	timeArray = []
	changeArray = []
	worksTotalTime = 0
	worksTimeCount = 0
	durationArray = []
	for works in worksDict.values():
		worksVid.append(works.vid)
		worksAid.append(works.subject.aid)
		matArray.append([len(works.playUids), works.like, len(works.shareFUids), len(works.shareIUids), len(works.replyPlayUids), len(works.longTimeUids), works.subAdd])
		timeArray.append([0 if works.timeCount == 0 else works.totalTime * 1.0 / works.timeCount])
		changeArray.append([0 if works.show == 0 else works.showPlay * 1.0 / works.show])
		worksTotalTime += works.totalTime
		worksTimeCount += works.timeCount
		durationArray.append([works.duration, len(works.playUids)])

	durationMat = mat(durationArray)
	durationTotal = durationMat[:,1].sum(axis=0)[0,0]
	durationAvg = multiply(durationMat[:,0], durationMat[:,1]).sum(axis=0)[0,0] / durationTotal
	durationVari = sqrt(multiply(square(durationMat[:,0] - durationAvg), durationMat[:,1]).sum(axis=0)[0,0] / durationTotal)
	durationRate = exp(-square(durationMat[:,0] - durationAvg) / 2 / square(durationVari)) * (1 / math.sqrt(2 * math.pi) / durationVari)
	#print(concatenate([durationRate, durationMat[:,0]], axis=1))

	worksTimeAvg = worksTotalTime * 1.0 / worksTimeCount

	timeMat = mat(timeArray) / worksTimeAvg
	changeMat = mat(changeArray)
	worksArgsMat = imatrix.inOne(mat(matArray))

	worksMat = mat(zeros((len(matArray), 3)))
	worksMat[:, 0] = worksArgsMat[:, 0] * 1 + worksArgsMat[:, 1] * 5 + worksArgsMat[:, 2] * 10 + worksArgsMat[:, 3] * 15 + worksArgsMat[:, 4] * 5
	worksMat[:, 1] = multiply(multiply(worksArgsMat[:, 5] * 10 + worksArgsMat[:, 6] * 20, timeMat), changeMat)

	rate = 0
	for index in range(1, 5):
		rate += (dau[index] - dau[index + 1]) * 1.0 / dau[index + 1]
	activeDiv = dau[1] * (1 + rate / 5) * 0.4
	a = 1 / (5.5 * activeDiv)
	factor = min(10, max(1, 1 / (a * dau[0])))

	worksMat = imatrix.inOne(worksMat)
	worksMatFinal = worksMat[:, 0] * factor + worksMat[:, 1] * (11 - factor)
	worksMatFinalOne = imatrix.inOne(worksMatFinal)

	subjectAids = []
	subjectArray = []
	for subject in subjectDict.values():
		subjectAids.append(subject.aid)
		subjectArray.append(subject.getArgs()[1:])

	subjectMat = mat(subjectArray)
	subjectMatOne = imatrix.inOne(subjectMat)

	subjectMatFinal = multiply((subjectMatOne[:, 0] * 20 + subjectMatOne[:, 1] * 1 + subjectMatOne[:, 2] * 5), subjectMat[:, 3]) * durationRate[index, 0]
	subjectMatFinalOne = imatrix.inOne(subjectMatFinal)

	worksHots = []
	for index, aid in enumerate(worksAid):
		worksInfo = []
		worksInfo.append(worksVid[index])
		worksInfo.append((worksMatFinalOne[index, 0] + subjectMatFinalOne[subjectAids.index(aid), 0] * 0.3) * worksDict[worksVid[index]].getDecayRatio())
		worksInfo.append(worksMatFinalOne[index, 0])
		worksInfo.append(subjectMatFinalOne[subjectAids.index(aid), 0])
		worksInfo.append(worksDict[worksVid[index]].getDecayRatio())
		worksInfo.append(durationRate[index, 0])
		worksInfo = worksInfo + worksDict[worksVid[index]].getArgs()[1:]
		worksHots.append(worksInfo)
	print(len(worksHots))
	return worksHots

	'''


	#! /usr/bin/env python
# -*- coding: utf-8 -*-
import time, sys, MySQLdb, redis
from elasticsearch import Elasticsearch

esTimeFormat = u"%Y-%m-%dT%H:%M:%S.%fZ"
esTimeFormatNoMs = u"%Y-%m-%dT%H:%M:%SZ"

testEsclient = Elasticsearch(['119.23.69.206:9200'], timeout=60)
devEsclient = Elasticsearch(['192.168.0.100:9200'], timeout=60)
localEsclient = Elasticsearch(['127.0.0.1:9200'], http_auth=('elastic', 'changeme'), timeout=10)
dataEsclient = Elasticsearch(['10.168.0.119:9200', '10.168.0.121:9200', '10.168.0.118:9200', '10.168.0.127:9200'], http_auth=('huangzhen', 'df25j0934j4y09hjoijoigef'), timeout=60)
relEsclient = Elasticsearch(['10.32.0.223:9200', '10.33.0.11:9200', '10.167.0.246:9200', '10.0.0.53:9200'], http_auth=('huangzhen', 'huangzhen'), timeout=60)


isTest = True

def setEnv(test):
	global isTest
	isTest = test

def updateRedisWorksHots(worksHots):
	r = redis.Redis(host='r-wz9536616f776d74.redis.rds.aliyuncs.com', port=6379, db=3, password='Idltest503') if isTest else redis.Redis(host='r-bp1f37095aef68e4.redis.rds.aliyuncs.com', port=6379, db=8, password='flrjovlOJEROIJ324')
	
	cacheKeyKey = 'recommend:srv:downgrade:keykey'
	cacheName = 'recommend:srv:downgrade:cache'
	backupCacheName = 'recommend:srv:downgrade:backupcache'

	#print(r.get(cacheKeyKey))
	if r.get(cacheKeyKey) == cacheName:
		r.zadd(backupCacheName, worksHots)
		r.expire(cacheName, 60)
		r.set(cacheKeyKey, backupCacheName)
	else:
		r.zadd(cacheName, worksHots)
		r.expire(backupCacheName, 60)
		r.set(cacheKeyKey, cacheName)

	return True

def searchDbConfig():
	config = {'hotVids': [], 'aidWeights': {}, 'vidWeights': {}}
	
	db = getDb()
	config['hotVids'] = [item[0] for item in searchDb(db, 'select vid from work_whitelist')]
	now = str(int(time.time()))
	for item in searchDb(db, 'select aid, weight from author_weight where start_time <= ' + now + ' and finish_time >= ' + now):
		config['aidWeights'][item[0]] = item[1]
	for item in searchDb(db, 'select vid, weight from work_weight where start_time <= ' + now + ' and finish_time >= ' + now):
		config['vidWeights'][item[0]] = item[1]
	db.close()

	return config

def searchDb(db, sql):
	cursor = db.cursor()
	cursor.execute(sql)
	return cursor.fetchall()

def getDb():
	path = "rm-bp12bzgmvo85rflhio.mysql.rds.aliyuncs.com" if isTest else "rm-bp10b4n5116u90h5t.mysql.rds.aliyuncs.com"
	username = "migration_test" if isTest else "webuser"
	password = "migrationTest*" if isTest else "fnwkufkljfk984ewJ"
	dbName = "idl_night" if isTest else "db_nt_operations"
	return MySQLdb.connect(path, username, password, dbName, charset='utf8')

def updateStatisticsAlias(index, isRead):
	indices = getEsclient().indices
	alias = 'statistics_' + index + '_' + ('read' if isRead else 'write')
	cIndex = indices.get_alias(index=alias).keys()[0]
	newIndex = cIndex[:len(cIndex)-8] if cIndex.endswith('_backups') else cIndex + '_backups'
	body = {
		"actions": [
			{"remove": {"index": cIndex, "alias": alias}},
			{"add": {"index": newIndex, "alias": alias}}
		]
	}
	indices.update_aliases(body=body)

def coverStatisticsData(index, dataDict):
	#先将读别名指向备份，再写入主索引，再将读别名指回
	updateStatisticsAlias(index, True)
	success = coverStatisticsDataReal(index, dataDict)
	#
	time.sleep(20)
	updateStatisticsAlias(index, True)
	if not success:
		return False

	#将写别名指向备份，再写入备份，再将写别名指回
	updateStatisticsAlias(index, False)
	success = coverStatisticsDataReal(index, dataDict)
	updateStatisticsAlias(index, False)
	return success

def coverStatisticsDataReal(index, dataDict):
	fullIndex = 'statistics_' + index + '_write'
	hits = scrollSearch(fullIndex, 'data', {"size": 1}, False)
	updateMark = len(hits) == 0 or not (hits[0]['_source']['updateMark'] if hits[0]['_source'].has_key('updateMark') else False)
	for item in dataDict.values():
		item['updateMark'] = updateMark
	if updateStatisticsData(index, dataDict):
		#操作频繁会发生冲突
		time.sleep(20)
		return deleteStatisticsData(fullIndex, {"query":{"bool":{"must_not":{"term":{"updateMark": updateMark}}}}})
	else:
		return False

def clearStatisticsData(index):
	deleteStatisticsData('statistics_' + index + '_write', {"query":{"match_all":{}}})

def deleteStatisticsData(fullIndex, body, doc_type='data'):
	success = len(getEsclient().delete_by_query(index=fullIndex, doc_type=doc_type, body=body, conflicts='proceed', preference='_primary_first')['failures']) == 0
	getEsclient().delete_by_query(index=fullIndex, doc_type=doc_type, body=body, conflicts='proceed')
	return success

def updateStatisticsData(index, dataDict):
	doc = []
	for key, value in dataDict.items():
		doc.append({"index":{"_id":key}})
		doc.append(value)
	return updateStatisticsDoc(index, doc)

def updateStatisticsDoc(index, doc):
	print('update', index, len(doc) / 2)
	start = 0
	while start < len(doc):
		#sys.stdout.write(str(start) + '\r')
		#sys.stdout.flush()
		end = min(start + 500, len(doc))
		success = updateStatisticsDocResolve(index, doc[start:end])
		if success:
			start = end
	return True

def updateStatisticsDocResolve(index, doc):
	#hot list额外在开发环境存一份用于测试
	#if isTest and index == 'works_hot':
	#	esUpdateStatisticsDocResolve(devEsclient, index, doc)
	success = esUpdateStatisticsDocResolve(getEsclient('statistics_' + index + '_write'), index, doc)
	#正式环境works_actions存一份在本地es用于离线训练
	try:
		if not isTest and index == 'works_actions':
			esUpdateStatisticsDocResolve(localEsclient, index, doc)
	except Exception as error:
		print('save local es error: ' + str(error))

	return success


def esUpdateStatisticsDocResolve(es, index, doc, doc_type='data'):
	return not es.bulk(index='statistics_' + index + '_write', doc_type=doc_type, body=doc)['errors']

def searchStatisticsAggs(index, body):
	fullIndex = 'statistics_' + index + '_read'
	return getEsclient(fullIndex).search(index=fullIndex, doc_type='data', body=body)['aggregations']

def searchPersonas(lastPersonasTime):
	body = {
	  "query":{
	    "bool": {
	      "must": [
	        {
	          "term": {
	            "type": "personas"
	          }
	        },
	        {
	          "range": {
	            "create_time": {
	              "gt": timeToEsTime(lastPersonasTime),
	              "lte": timeToEsTime(min(lastPersonasTime + 12 * 60 * 60, int(time.time()) - 90))
	            }
	          }
	        }
	      ]
	    }
	  },
  		"size": 10000
	}
	return scrollSearch('clientlog_read', 'watch', body)

def searchStatisticsData(index, size=10000, searchAll=True, log=True, version=False):
	return searchStatisticsDataByBody(index, {"size": size}, searchAll, log, version)

def searchStatisticsDataByBody(index, body, searchAll=True, log=True, version=False):
	return scrollSearch('statistics_' + index + '_read', 'data', body, searchAll, log, version)

def scrollSearch(fullIndex, doc_type, body, searchAll=True, log=True, version=False):
	esclient = getEsclient(fullIndex)
	hits = []
	if searchAll:
		result = esclient.search(index=fullIndex, doc_type=doc_type, scroll='1m', body=body, version=version, preference='_primary_first')
		while len(result['hits']['hits']) != 0:
			hits.extend(result['hits']['hits'])
			scroll_id = result['_scroll_id']
			result = esclient.scroll(scroll_id=scroll_id, scroll='1m')
	else:
		hits.extend(esclient.search(index=fullIndex, doc_type=doc_type, body=body, version=version, preference='_primary_first')['hits']['hits'])
	if log:
		print('get', fullIndex, len(hits))
	return hits

def getEsclient(fullIndex=None):
	return testEsclient if isTest else relEsclient if fullIndex == 'works_video_read' else localEsclient if fullIndex == 'statistics_works_kartuns_write' else dataEsclient

def timeToEsLocalTime(pTime):
	return time.strftime('%Y-%m-%dT%H:%M:%S+0800', time.localtime(pTime))

def timeToEsTime(pTime, hasMs=False):
	return time.strftime(esTimeFormat if hasMs else esTimeFormatNoMs, time.gmtime(pTime))

def esTimeToTime(esTime, hasMs=False):
	return time.mktime(time.strptime(esTime, esTimeFormat if hasMs else esTimeFormatNoMs)) + 28800


